/* riva_iic.c PCI IIC DRIVER V0.23 3-2002
** written by Rudolf Cornelissen
** Email: rag.cornelissen@inter.nl.net
*/

//keep 'X86/_KERNEL_' added in project!!

#define VERSION 23		//driver version number = 23/100 = 0.23
#define APPVERSION 118	//app version number required = 0.118

#include <KernelExport.h>
#include <Drivers.h>
#include <ISA.h> //leave it inhere for testing purposes please...
#include <PCI.h>
#include <malloc.h>

#include "riva_iic.h"

//coldstart use:
typedef struct {
	uint8	ram,
			wr_delay;
} strapstruct;
//end coldstart use.

typedef struct rivapci_private {
	int32			devID; 		/* device identifier: 0-n */
	pci_info		*pciInfo;
	uint32			reg_base,	/* hardware register base address */
					size;		/* hardware register area size */
	area_id			ioarea, 	/* Area used for MMIO of hostRegs */
					fbarea,		/* framebuffer area */
					romarea;	/* BIOS ROM area */
	vuint32			*regs,		/* kernel-memory hardware registers base adress (for 32bit access)*/
					*fbout,		/* kernel-memory outbound framebuffer base adress (for 32bit access)*/
					*fbin;		/* kernel-memory inbound frambuffer base adress for 32bit access)*/
	vuint8			*pcio_base,	/* 'adress $000' in conventional I/O space (for 8bit access)*/
					*pdio_base;	/* ehh.. later.. */
} rivapci_private_t;

//Pitfalls about pointers to (PCI) memory:
//Make sure:
//- You use pointers to 8bit 'variables' to read or write 8bits at a time from/to PCI;
//- You use pointers to 16bit 'variables' to read or write 16bits at a time from/to PCI;
//- You use pointers to 32bit 'variables' to read or write 32bits at a time from/to PCI;
//
//Also make sure these 'variables' are vuint8, vuint16 or vuint32. If you use a 'char' instead
//of a 'vuint8' for example it will work most of the time. However, with some registers this
//approach will fail.
//The difference here in these types of 'variables' has to do with the useage in hardware of the
//chip-selection lines like READ, WRITE, IOSEL, MEMSEL. Also a 'caching' difference may exist.
//
//Registers that are read for instance to *not* read their contents, but purely for resetting
//some other register depends on this difference for success!
//
//EXAMPLE: the register at offset 0x3da or 0x6013da (VGA input status #1).
//This one has two functions.
//1. It's used to read the VGA's input status #1 register, which contains info about being in
//   vertical retrace or not;
//2. It's also used by purely accessing this register with a READ command (so read contents may
//   be thrown away) for selecting the Attribute controller registers ADRESS/INDEX register.

static int32    get_pci_list(pci_info *info[], int32 maxEntries);  /* Get pci_info for each device */
static status_t free_pci_list(pci_info *info[]);                   /* Free storage used by pci_info list */
static status_t enable_addressing(rivapci_private_t  *data);       /* enable pci io address space for device */
void RMA_PCI_Write(rivapci_private_t *dev, uint32 adress, uint32 data);
uint32 RMA_PCI_Read(rivapci_private_t *dev, uint32 adress);
void testRAM(rivapci_private_t *dev);
void SwitchToVGA(rivapci_private_t *dev);
void SwitchRIVAtoTV (char lsb, char msb, char type, rivapci_private_t *dev);
void EnableTVOut(rivapci_private_t *dev);
void OutSCL(char Bit, int BusNR, rivapci_private_t *dev);
void OutSDA(char Bit, int BusNR, rivapci_private_t *dev);
char InSCL(int BusNR, rivapci_private_t *dev);
char InSDA(int BusNR, rivapci_private_t *dev);
void bstart (int BusNR, rivapci_private_t *dev);
void bstop (int BusNR, rivapci_private_t *dev);
void TXBit (char Bit, int BusNR, rivapci_private_t *dev);
char RXBit (int BusNR, rivapci_private_t *dev);
void TXByte (char Byte, int BusNR, rivapci_private_t *dev);
char RXByte (char Ack, int BusNR, rivapci_private_t *dev);
char FlagIICError (char ErrNo);

#define MAX_CARDS 			 1			/* maximum number of driver instances */

static char 				*gDevNameList[MAX_CARDS+1];
static pci_info 			*gDevList[MAX_CARDS+1];



//version control: following line just should be here..
int32 api_version = B_CUR_DRIVER_API_VERSION;

//used NVIDIA TNT/TNT2/GFORCE/GFORCE2 registers: (are fixed at these I/O adresses.)
const unsigned int 	AdresReg = 0x03d4,
					DataReg = 0x03d5;

//used index adresses within NVIDIA chipset:
const unsigned char RDIICBus0 = 0x3e,
					WRIICBus0 = 0x3f,
					RDIICBus1 = 0x36,
					WRIICBus1 = 0x37;

/*
** Bit definition of RDIICBusX: (do not write here!)
** bit7: unknown
** bit6: unknown
** bit5: unknown
** bit4: unknown
** bit3: SDA line input
** bit2: SCL line input
** bit1: unknown
** bit0: unknown
*/

/*
** Bit definition of WRIICBusX:
** bit7: don't change
** bit6: don't change
** bit5: SCL line output (R/W)
** bit4: SDA line output (R/W)
** bit3: write '0'
** bit2: write '0'
** bit1: write '0'
** bit0: write '1'
*/

static spinlock lock = 0;
#define ENTER_CRITICAL_SECTION { cpu_status former; former = disable_interrupts(); acquire_spinlock(&lock);
#define LEAVE_CRITICAL_SECTION release_spinlock(&lock); restore_interrupts(former); }

isa_module_info* ISA_Bus;
pci_module_info* PCI_Bus;
ushort vendorID, deviceID;
//vuint32 *FBout = NULL, *FBin = NULL;

status_t
init_hardware (void)
{
	dprintf("riva_iic: driver hardware init called.\n");
	return(B_OK);
}


status_t
init_driver (void)
{
	status_t ret;
	int32 entries;
	int i;
	char devName[64] = "video/riva_iic/1";

//ISA:
  	ret = get_module (B_ISA_MODULE_NAME,(module_info**)&ISA_Bus);
	if (ret<0)
	{
		dprintf ("\nriva_iic: driver init returns negative\n");
		return (ENODEV);
	}


  	ret = get_module (B_PCI_MODULE_NAME,(module_info**)&PCI_Bus);
	if (ret<0)
	{
		dprintf ("\nriva_iic: no PCI bus found\n");
		return (ENODEV);
	}
	else
	{

		/* Find RIVA cards*/
		if ((entries = get_pci_list(gDevList, MAX_CARDS ))  == 0)
	 	{
			dprintf ("\nriva_iic: no RIVA videocard devices found\n");
			free_pci_list(gDevList);
			put_module(B_PCI_MODULE_NAME);
			return (ENODEV);
		}

		/* Create device name list*/
		for (i=0; i<entries; i++ )
		{
			//sprintf(devName, "%s%ld", kDevDir, i );
			gDevNameList[i] = (char *)malloc(strlen(devName)+1);
			strcpy(gDevNameList[i], devName);
		}
		gDevNameList[i] = NULL;

		dprintf("\nriva_iic: driver init returns OK\n");
		
		return(B_OK);
	}
}

void
uninit_driver (void)
{
	int32 	i;
	void 	*item;

	/*Free device name list*/
	for (i=0; (item=gDevNameList[i]); i++)
		free(item);

	/*Free device list*/
	free_pci_list(gDevList);
	put_module(B_ISA_MODULE_NAME);
	put_module(B_PCI_MODULE_NAME);
	dprintf("riva_iic: driver uninited\n");
}	

static int32 get_pci_list(pci_info *info[], int32 maxEntries)
{
	status_t status;
	int32 i, entries;
	pci_info	*item;
	
	item = (pci_info *)malloc(sizeof(pci_info));
	
	for (i=0, entries=0; entries<maxEntries; i++)
	{
		if ((status = PCI_Bus->get_nth_pci_info(i, item)) != B_OK) break;
		if (item->vendor_id == 0x10de)	// NVIDIA

/*		if (   ((item->vendor_id == 0x10de) && (item->device_id == 0x0028))		// NVIDIA TNT2
			|| ((item->vendor_id == 0x10de) && (item->device_id == 0x0110)) 	// NVIDIA GeForce2 MX200
			|| ((item->vendor_id == 0x10de) && (item->device_id == 0x002d)) 	// NVIDIA TNT2-M64
			|| ((item->vendor_id == 0x10de) && (item->device_id == 0x0020))	)	// NVIDIA TNT1
*/		{
			dprintf("riva_iic: RIVA videocard found; vendor_id 0x%x, device_id 0x%x;\n",
																item->vendor_id, item->device_id);
			dprintf("riva_iic: bus: %d, device: %d, function %d\n",
														item->bus, item->device, item->function);
			//report these to the user later on:
			vendorID = item->vendor_id;
			deviceID = item->device_id;
			info[entries++] = item;
			item = (pci_info *)malloc(sizeof(pci_info));
		}
	}
	info[entries] = NULL;
	free(item);
	return entries;
}
	
static status_t free_pci_list(pci_info *info[])
{
	int32 		i;
	pci_info	*item;
	
	for (i=0; (item=info[i]); i++)
		free(item);
	return B_OK;
}

static status_t enable_addressing(rivapci_private_t  *dev)
{
	uint32 cmd;//, offset, size, base;
	char buffer[B_OS_NAME_LENGTH];

	dev->reg_base = dev->pciInfo->u.h0.base_registers_pci[0];//control regs viewed from PCI space
	dev->size = dev->pciInfo->u.h0.base_register_sizes[0];
	dprintf("riva_iic: card reg_base: 0x%x\n", (uint) dev->reg_base);
	dprintf("riva_iic: card ioarea_size: 0x%x\n", (uint) dev->size);

	dprintf("riva_iic: card fb_base: 0x%x\n",
		(uint) dev->pciInfo->u.h0.base_registers_pci[1]);
	dprintf("riva_iic: card fb_size: 0x%x\n",
		(uint) dev->pciInfo->u.h0.base_register_sizes[1]);

	/* enable pci address access */	
	cmd = (PCI_Bus->read_pci_config)(dev->pciInfo->bus, dev->pciInfo->device, dev->pciInfo->function, PCI_command, 4);

	cmd |= PCI_command_memory;// clearing this bit hangs video: disables PCI access..
	cmd |= PCI_command_master;//VESA: no influence

//coldstart test:
	cmd |= PCI_command_io; //enables conventional ISA I/O space access..
//end coldstart test.

	//cmd &= ~PCI_command_mwi;			//VESA: no influence
	//cmd &= ~PCI_command_address_step;	//VESA: no influence
	//cmd &= ~PCI_command_vga_snoop;	//VESA: no influence
	//cmd &= ~PCI_command_serr;			//VESA: no influence
	//cmd &= ~PCI_command_parity;		//VESA: no influence
	//cmd &= ~PCI_command_fastback;		//VESA: no influence
	//cmd &= ~PCI_command_special;		//VESA: no influence

	(PCI_Bus->write_pci_config)(dev->pciInfo->bus, dev->pciInfo->device, dev->pciInfo->function, PCI_command, 4, cmd);

	//map card range in physical memory:
/*	base = data->pciInfo->u.h0.base_registers[0];
	size = data->pciInfo->u.h0.base_register_sizes[0];
*/	
	/* Round down to nearest page boundary */
//	data->reg_base = data->reg_base & ~(B_PAGE_SIZE-1);
	
	/* Adjust the size */
/*	offset = data->pciInfo->u.h0.base_registers[0] - data->reg_base;
	size += offset;
	size = (size +(B_PAGE_SIZE-1)) & ~(B_PAGE_SIZE-1);
*/
	//map control registers in memory:
	dev->ioarea = map_physical_memory(
		buffer,
		(void *) dev->reg_base,
		dev->size,/*size,*/
		B_ANY_KERNEL_ADDRESS,
		0, /* B_READ_AREA + B_WRITE_AREA, */ /* neither read nor write, to hide it from user space apps */
		(void **)&(dev->regs));
	/* return an error if there was some problem */
	if (dev->ioarea < 0) return B_ERROR;

	//map outbound framebuffer in memory:
	dev->fbarea = map_physical_memory(
		buffer,
		(void *) dev->pciInfo->u.h0.base_registers_pci[1],
		dev->pciInfo->u.h0.base_register_sizes[1],
		B_ANY_KERNEL_ADDRESS,
		B_READ_AREA + B_WRITE_AREA,   /* accessible from user space */
		(void **)&(dev->fbout));
	/* return an error if there was some problem */
	if (dev->fbarea < 0)
	{
		//unmap already mapped part:
		delete_area(dev->ioarea);
		dev->ioarea = -1;
		return B_ERROR;
	}
	//set inbound framebuffer:
	dev->fbin = NULL;

	dprintf("riva_iic: kernel-memory reg_base: 0x%x\n", (uint) dev->regs);
	dev->pcio_base = (void *)((uint32)dev->regs + 0x00601000); //offset to get to PCIO-base

	dprintf("riva_iic: kernel-memory pcio_base: 0x%x\n", (uint) dev->pcio_base);

	return B_OK;
}

static status_t
iic_device_open (const char *name, uint32 flags, void** cookie)
{
	int32 devID;
	status_t status;
	char *devName;
	rivapci_private_t *dev;

	/*	Find device name*/
	for (devID=0; (devName=gDevNameList[devID]); devID++) {
		if (strcmp(name, devName) == 0)
			break;
	}
	if (!devName)
		return (B_ERROR);

	/*	Allocate storage for the cookie*/
	if (!(*cookie = dev = (rivapci_private_t *)malloc(sizeof(rivapci_private_t))))
	{
		dprintf("riva_iic: no memory available\n");
		return(B_NO_MEMORY);
	}
	memset(dev, 0, sizeof(rivapci_private_t));
	
	/* Setup the cookie */
	dev->pciInfo = gDevList[devID];
	dev->devID = devID;

	/* enable access to the cards address space */
	if ((status = enable_addressing(dev)) != B_OK)
	{	
		dprintf("riva_iic: access to cards adress space failed\n");
		return(status);
	}
	else
	{
		dprintf("riva_iic: device opened\n");
		return(B_OK);
	}
}


static status_t
iic_device_read (void* cookie, off_t position, void *buffer, size_t* num_bytes)
{
	//A way should be found to block others from entering here if a IIC bus
	//is busy for another! (bus is busy between first START and first STOP condition.)

	rivapci_private_t *dev = (rivapci_private_t *) cookie;	//internal device info
	riva_data_t *data = (riva_data_t *) buffer;				//user device data

	int bytes = *num_bytes;
	int count;
	char* iicdata = (char*)data->iic.buffer;

	//uint32 temp;

	dprintf("riva_iic: reading: %d byte(s) on IIC bus %d\n",bytes,data->iic.busnr);

	for (count = 0; count < bytes; count++)
		*iicdata++ = RXByte (*iicdata,data->iic.busnr, dev);

	//reset pointer to bufferstart!
	iicdata = (char*)data->iic.buffer;
	dprintf("riva_iic: data is:\n");
	for (count = 0; count < bytes; count++)
		dprintf("%d -> 0x%x\n", count, (int)*iicdata++);
	dprintf(".\n");

/*
	//32 bit PCI access:
	temp = *(dev->regs + (0x0068050c >> 2));
	rxdata[1] = temp;
	rxdata[2] = temp >> 8;
	rxdata[3] = temp >> 16;
	rxdata[4] = temp >> 24;
			
	//8 bit PCI access:			
	rxdata[5] = *(dev->pcio_base + (0x0007f50c));
	rxdata[6] = *(dev->pcio_base + (0x0007f50d));
	rxdata[7] = *(dev->pcio_base + (0x0007f50e));
	rxdata[8] = *(dev->pcio_base + (0x0007f50f));

	rxdata[0] = 0xff;//test
	memcpy(((char*)buffer),rxdata,20);
*/
	return (B_OK);
}


static status_t
iic_device_write (void* cookie, off_t position, const void* buffer, size_t* num_bytes)
{
	//A way should be found to block others from entering here if a IIC bus
	//is busy for another! (bus is busy between first START and first STOP condition.)

	rivapci_private_t *dev = (rivapci_private_t *) cookie;	//internal device info
	riva_data_t *data = (riva_data_t *) buffer;				//user device data

	int bytes = *num_bytes;
	int count;
	char* iicdata = (char*)data->iic.buffer;
	
	dprintf("riva_iic: writing: %d byte(s) on IIC bus %d\nriva_iic: data is:",bytes,data->iic.busnr);
	for (count = 0; count < bytes; count++)
		dprintf("%d -> 0x%x\n", count, (int)*iicdata++);
		
	dprintf(".\n");

	//reset pointer to bufferstart!
	iicdata = (char*)data->iic.buffer;
	for (count = 0; count < bytes; count++)
		TXByte (*iicdata++,data->iic.busnr, dev);
	
	return (B_OK);
}


static status_t
iic_device_control (void* cookie, uint32 operation, void* args, size_t length)
{
	rivapci_private_t *dev = (rivapci_private_t *) cookie;	//internal device info
	riva_data_t *data = (riva_data_t *) args;				//user device data

	int err = B_OK;
	unsigned char lsb, msb, VertRetrace;
	uint32 temp32,
		 index32; //test for fillpattern generation...
	uint8 index; //for setmode...

	dprintf("riva_iic: device control (%d/0x%x) called\n", (int)operation, (int)operation);
 
    switch (operation)
    { 
    case rivaStartIIC:	
    		dprintf("riva_iic: generate start condition on IIC bus %d\n",data->iic.busnr);  
			FlagIICError (-1);//clear IIC errorcode
			//A way should be found to block others from entering here if a IIC bus
			//is busy! (bus is busy between first START and first STOP condition.)
			bstart(data->iic.busnr, dev);
			if (FlagIICError (0)) err = B_IO_ERROR;
			//return IIC err code also...
			data->iic.status = FlagIICError (0);
			//do not reset IICError, first error on bus is the valid one..
            break; 
	case rivaStopIIC:
			dprintf("riva_iic: generate stop condition on IIC bus %d\n",data->iic.busnr);
			bstop(data->iic.busnr, dev);
			//A way should be found to disable the block started on the first
			//START condition! (bus is free now.)
			if (FlagIICError (0)) err = B_IO_ERROR;
			//return IIC err code also...
			data->iic.status = FlagIICError (0);
			FlagIICError (-1);//clear IIC errorcode
			break;
	case rivaChkBus:
			//A way should be found to block others from entering here if a IIC bus
			//is busy! (bus is busy between first START and first STOP condition.)

			spin(6);
			bstop(data->iic.busnr, dev);//init and/or stop comms IIC bus
			dprintf("riva_iic: check for existance of IIC bus %d: ",data->iic.busnr);
			//check for hardware coupling of SCL and SDA -out and -in lines.
			spin(6);
			OutSCL(0,data->iic.busnr, dev);
			OutSDA(1,data->iic.busnr, dev);
			spin(3);
			if (InSCL(data->iic.busnr, dev) || !InSDA(data->iic.busnr, dev))
			{
				err = B_IO_ERROR;
				dprintf("fail.\n");
				break;
			}
			spin(3);
			OutSCL(1,data->iic.busnr, dev);
			OutSDA(0,data->iic.busnr, dev);
			spin(3);
			if (!InSCL(data->iic.busnr, dev) || InSDA(data->iic.busnr, dev))
			{
				err = B_IO_ERROR;
				dprintf("fail.\n");
				break;
			}
			dprintf("OK.\n");
			spin(3);
			bstop(data->iic.busnr, dev);//re-init IIC bus
			break;
	case rivaVersion:
			if (data->drv_version != 0xff)
			{
				data->drv_version = VERSION;
				dprintf("riva_iic: app failed to unlock me; stopping it...\n");
            	err = B_DEV_INVALID_IOCTL;
				break;
			}
			data->drv_version = VERSION;
			if (data->app_version < APPVERSION)
			{
				//report needed appversion...
				data->app_version = (uint16) APPVERSION;
				dprintf("riva_iic: app too old for me; stopping it...\n");
            	err = B_DEV_INVALID_IOCTL;
				break;
			}
			dprintf("riva_iic: return driver version nr.: %d/100\n",data->drv_version);
			//fill out data structure for user to use:
			data->dev.vendor_id = vendorID;
			data->dev.device_id = deviceID;
			data->fb.fbmout = dev->fbout;
			data->fb.fbmin = dev->fbin;
			data->analog.outfield = 0;
			data->fb.fbout_sw = 0;
			//not yet known how to do this:
			data->fb.fbmout_size = 0;
			data->fb.fbmin_size = 0;
			
			//get currently used framebuffer baseadress (window manager):
			//b1-0:
			//(select VGA ATTR index register:)
			*(dev->pcio_base + 0x3da);
			//(select ATTR register to read:)
			*(dev->pcio_base + 0x3c0) = 0x33;
			//(read data:)
			temp32 = ((*(dev->pcio_base + 0x3c1) & 0x06) >> 1);
			//read b2-9 from SA_LO:
			*(dev->pcio_base + AdresReg) = 0x0d;
			temp32 |= (*(dev->pcio_base + DataReg) << 2);
			//read b17-10 from SA_HI:
			*(dev->pcio_base + AdresReg) = 0x0c;
			temp32 |= (*(dev->pcio_base + DataReg) << 10);
			//read b22-18 from repaint0:
			*(dev->pcio_base + AdresReg) = 0x19;
			temp32 |= ((*(dev->pcio_base + DataReg) & 0x1f) << 18);
			//read b24-23 from HEB:
			*(dev->pcio_base + AdresReg) = 0x2d;
			temp32 |= ((*(dev->pcio_base + DataReg) & 0x60) << 18);
			data->fb.fbw = (vuint32*)temp32;
			
			//currently displayed framebuffer is the fbw buffer:
			data->fb.fbout_used = 2;

			break;
    case rivaUnlock:
    		dprintf("riva_iic: unlocking extended registers for R/W\n");
			ENTER_CRITICAL_SECTION

			//NV_PMC_ENABLE: enable all nvidia hardware parts:
			*(dev->regs + (0x00000200 >> 2)) = 0x11111111;
			//PVIDEO disable: VESA: no influence
			//PCRTC disable: disables PCI memaccess also.. (video gone)
			//PFB disable: disables frame buffer access,
				// so the last accessed line keeps repeating (in effect: video gone)
			//PPMI disable: VESA: no influence
			//PGRAPH disable: VESA: no influence
			//PFIFO disable: VESA: no influence
			//PMEDIA disable: VESA: no influence
			//bit0 disable: powers down add-on chips on GeForce and higher cards!

		//coldstart test:
		//setup VIO_VSE2 register:
		*(dev->pcio_base + 0x3c3) = 0x01;
		//(MISC.OUTPUT: set color mode: determines most std. VGA register locations!)
		*(dev->pcio_base + 0x3c2) = ((*(dev->pcio_base + 0x3cc)) | 0x01);
		//end coldstart test.
		
			//unlock registers:
			*(dev->pcio_base + AdresReg) = 0x1f;
			*(dev->pcio_base + DataReg) = 0x57;//write 'magic number'
 			LEAVE_CRITICAL_SECTION
    		break;
    case rivaPal640:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) PAL 50Hz 640x480\n");
			SwitchRIVAtoTV (0x39, 0x02, data->type, dev);
	   		break;
    case rivaPal800:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) PAL 50Hz 800x600\n");
			SwitchRIVAtoTV (0xb6, 0x02, data->type, dev);
	   		break;
    case rivaNtsc640:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) NTSC 60Hz 640x480\n");
			SwitchRIVAtoTV (0x2a, 0x02, data->type, dev);
	   		break;
    case rivaNtsc800:
    		dprintf("riva_iic: switch RIVA chip for (dualhead clone) NTSC 60Hz 800x600\n");
			SwitchRIVAtoTV (0xac, 0x02, data->type, dev);
	   		break;
	case rivaEnableTVOut:
			dprintf("riva_iic: switch RIVA chip for tv out\n");
			EnableTVOut(dev);
			break;
	case rivaVgaOff:
			dprintf("riva_iic: turn off VGA monitor\n");
			//REPAINT1: switch off VGA monitor HSYNC
			ENTER_CRITICAL_SECTION
			*(dev->pcio_base + AdresReg) = 0x1a;
			*(dev->pcio_base + DataReg) |= 0x80;
 			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVgaOn:
			dprintf("riva_iic: turn on VGA monitor\n");
			//REPAINT1: switch on VGA monitor HSYNC
			ENTER_CRITICAL_SECTION
			*(dev->pcio_base + AdresReg) = 0x1a;
			*(dev->pcio_base + DataReg) &= 0x7f;
 			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVesaVgaOnly:
			dprintf("riva_iic: switch RIVA chip for (VESA) VGA only\n");
			//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869)
			//and Diamond Viper V550 16Mb PCI (TNT1/BT869)

			//works OK if booted in VESA on VGA or TV;
			//(sets refreshrate = 60Hz).
			//(for non VESA VGA the BScreen API and 'rivaVGAOnly' should be used.)

			//switch RIVA to VGA:
			SwitchToVGA(dev);

			//set RIVA picture timing registers (VESA setup used):
			//read hor display end:
			ENTER_CRITICAL_SECTION
			*(dev->pcio_base + AdresReg) = 0x01;
			if (*(dev->pcio_base + DataReg) == 0x63)
			{
				//800x600 resolution active...
				lsb = 0x72;//values for vertical total...
				msb = 0x02;

				//values for RIVA pixelclock (60Hz):
				temp32 = 0x00027b0b;

				//set hor blank start:
				*(dev->pcio_base + AdresReg) = 0x02;
				*(dev->pcio_base + DataReg) = 0x63;

				//set hor restrace start:
				*(dev->pcio_base + AdresReg) = 0x04;
				*(dev->pcio_base + DataReg) = 0x6a;

				//set hor display total:
				*(dev->pcio_base + AdresReg) = 0x00;
				*(dev->pcio_base + DataReg) = 0x7f;

				//set hor retrace end:
				*(dev->pcio_base + AdresReg) = 0x05;
				*(dev->pcio_base + DataReg) = 0x1a;

				//set hor blank end:
				*(dev->pcio_base + AdresReg) = 0x03;
				*(dev->pcio_base + DataReg) = 0x83;

				//set LSR:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vertical display end:
				*(dev->pcio_base + AdresReg) = 0x12;
				*(dev->pcio_base + DataReg) = 0x57;

				//set vert blank start:
				*(dev->pcio_base + AdresReg) = 0x15;
				*(dev->pcio_base + DataReg) = 0x57;

				//set overflow:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0xf0;

				//set height:
				*(dev->pcio_base + AdresReg) = 0x09;
				*(dev->pcio_base + DataReg) = 0x60;

				//unlock VGA vertical retrace start and end registers for PCI access!
				//(nessesary here per se for PCI access:
				// access expires on writing to other registers than just those two...)
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) &= 0x7f;
				*(dev->pcio_base + AdresReg) = 0x1f;
				*(dev->pcio_base + DataReg) = 0x57;

				//unlock VGA registers 'vert. retrace start' & 'vert. retrace end' for ISA access:
				//(may be placed 'everywhere' up to this point in the code:
				// access stays enabled until locked manually again...)
//				ISA_Bus->write_io_8(AdresReg,0x03);
//				ISA_Bus->write_io_8(DataReg,
//					(ISA_Bus->read_io_8(DataReg) | 0x80));

				//set vert retrace start:
				*(dev->pcio_base + AdresReg) = 0x10;
				*(dev->pcio_base + DataReg) = 0x59;

				//set overflow vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0xf0;

				//set LSR vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vert. retrace end:
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) = 0x0d;

				//set vert. blank end:
				*(dev->pcio_base + AdresReg) = 0x16;
				*(dev->pcio_base + DataReg) = 0x73;
			}
			else
			{
				//assume 640x480 resolution active...
				lsb = 0x0b;//values for vertical total...
				msb = 0x02;

				//values for RIVA pixelclock (60Hz):
				temp32 = 0x0003c50e;

				//set hor blank start:
				*(dev->pcio_base + AdresReg) = 0x02;
				*(dev->pcio_base + DataReg) = 0x4f;

				//set hor restrace start:
				*(dev->pcio_base + AdresReg) = 0x04;
				*(dev->pcio_base + DataReg) = 0x53;

				//set hor display total:
				*(dev->pcio_base + AdresReg) = 0x00;
				*(dev->pcio_base + DataReg) = 0x5f;

				//set hor retrace end:
				*(dev->pcio_base + AdresReg) = 0x05;
				*(dev->pcio_base + DataReg) = 0x9f;

				//set hor blank end:
				*(dev->pcio_base + AdresReg) = 0x03;
				*(dev->pcio_base + DataReg) = 0x83;

				//set LSR:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vertical display end:
				*(dev->pcio_base + AdresReg) = 0x12;
				*(dev->pcio_base + DataReg) = 0xdf;

				//set vert blank start:
				*(dev->pcio_base + AdresReg) = 0x15;
				*(dev->pcio_base + DataReg) = 0xdf;

				//set overflow:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0x3e;

				//set height:
				*(dev->pcio_base + AdresReg) = 0x09;
				*(dev->pcio_base + DataReg) = 0x40;

				//unlock VGA vertical retrace start and end registers for PCI access!
				//(nessesary here per se for PCI access:
				// access expires on writing to other registers than just those two...)
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) &= 0x7f;
				*(dev->pcio_base + AdresReg) = 0x1f;
				*(dev->pcio_base + DataReg) = 0x57;

				//unlock VGA registers 'vert. retrace start' & 'vert. retrace end' for ISA access:
				//(may be placed 'everywhere' up to this point in the code:
				// access stays enabled until locked manually again...)
//				ISA_Bus->write_io_8(AdresReg,0x03);
//				ISA_Bus->write_io_8(DataReg,
//					(ISA_Bus->read_io_8(DataReg) | 0x80));

				//set vert retrace start:
				*(dev->pcio_base + AdresReg) = 0x10;
				*(dev->pcio_base + DataReg) = 0xea;

				//set overflow vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0x3e;

				//set LSR vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vert. retrace end:
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) = 0x0c;

				//set vert. blank end:
				*(dev->pcio_base + AdresReg) = 0x16;
				*(dev->pcio_base + DataReg) = 0x0c;
			}

			//Set the vertical total:
    		//LSB (vertical total):
			*(dev->pcio_base + AdresReg) = 0x06;
			*(dev->pcio_base + DataReg) = lsb;
			//b8 and b9:(overflow)
			*(dev->pcio_base + AdresReg) = 0x07;
			*(dev->pcio_base + DataReg) = 
				((*(dev->pcio_base + DataReg) & ~0x21) | ((msb & 0x01)+((msb & 0x02)<<4)));
			//b10:(LSR)
			*(dev->pcio_base + AdresReg) = 0x25;
			*(dev->pcio_base + DataReg) = 
				((*(dev->pcio_base + DataReg) & ~0x01) | ((msb & 0x04)>>2));

			//set RIVA pixelclock:
			//(MEMADR $00680508)
			*(dev->regs + (0x00680508 >> 2)) = temp32;
			LEAVE_CRITICAL_SECTION
	   		break;
	case rivaVgaOnly:
			dprintf("riva_iic: switch RIVA chip for (non-VESA) VGA only\n");
			//switch RIVA to VGA:
			SwitchToVGA(dev);
	   		break;
	case rivaIsMode640:
			dprintf("riva_iic: IsMode 640x480 check\n");
			ENTER_CRITICAL_SECTION
			//unlock VGA registers at index 0-7
			*(dev->pcio_base + AdresReg) = 0x11;
			*(dev->pcio_base + DataReg) &= 0x7f;
			//check hor display end for 640 pixels
			*(dev->pcio_base + AdresReg) = 0x01;
			if (!(*(dev->pcio_base + DataReg) == 0x4f)) err = B_IO_ERROR;
			LEAVE_CRITICAL_SECTION
			break;
	case rivaIsModeValid:
			dprintf("riva_iic: IsModeValid check\n");
			ENTER_CRITICAL_SECTION
			//unlock VGA registers at index 0-7
			*(dev->pcio_base + AdresReg) = 0x11;
			*(dev->pcio_base + DataReg) &= 0x7f;
			//check hor display end for 640 or 800 pixels
			*(dev->pcio_base + AdresReg) = 0x01;
			if (!(*(dev->pcio_base + DataReg) == 0x4f) &&
					!(*(dev->pcio_base + DataReg) == 0x63)) err = B_IO_ERROR;
			LEAVE_CRITICAL_SECTION
			break;
    case rivaVgaSync:	
    		dprintf("riva_iic: syncing to vertical retrace.\n");  
			//wait for start of VGA retrace...
			//(this is no critical section because only one register is only read!)
			//(make sure we are 'in' active VGA picture so we are sure to detect the
			// start of the vertical retrace moment:)
			do
			{
				VertRetrace = (*(dev->pcio_base + 0x3da) & 0x08);
				snooze(1000);
			} while (VertRetrace);
			//get the current fieldtype that is generated:
			//(not the correct register I'm afraid...)
			data->analog.outfield = 0;//((*(dev->regs + (0x00680700 >> 2)) & 0x10000000) >> 28);
			//now calculate the next one that will be started after retrace:
			data->analog.outfield = ((data->analog.outfield + 1) & 0x01); 
			//(wait for vertical retrace start on VGA:)
			do 
			{
				VertRetrace = (*(dev->pcio_base + 0x3da) & 0x08);
				snooze(100);
			} while (!VertRetrace);
            break; 
    case rivaGetMode:	
    		dprintf("riva_iic: GetMode() called.\n");  
			ENTER_CRITICAL_SECTION

			//read horizontal display end and translate:
			*(dev->pcio_base + AdresReg) = 0x01;
			data->displ_mode.displ_timing.h_display = ((*(dev->pcio_base + DataReg) + 1) * 8);

			//read vertical display end and translate:
		    //(LSB:)
			*(dev->pcio_base + AdresReg) = 0x12;
			temp32 = *(dev->pcio_base + DataReg);
			//b8 and b9:(overflow)
			*(dev->pcio_base + AdresReg) = 0x07;
			lsb = (*(dev->pcio_base + DataReg) & 0x42);
			//b10:(LSR)
			*(dev->pcio_base + AdresReg) = 0x25;
			msb = (*(dev->pcio_base + DataReg) & 0x02);
			//(combine all to temp32:)
			temp32 += ((((lsb & 0x02)>>1) + ((lsb & 0x40)>>5) + (msb<<1)) << 8);
			//translate:
			data->displ_mode.displ_timing.v_display = (temp32 + 1);
			
			LEAVE_CRITICAL_SECTION
			break;
    case rivaSetMode:	
    		dprintf("riva_iic: SetMode() called.\n");
    		//for now we will set the card to fixed 800x600x16 @ 60Hz: testcase..  
			ENTER_CRITICAL_SECTION
			//set card to color mode (sets registers to expected adresses)
			//misc. output:
			//(set color mode and enable RAM access)
			*(dev->pcio_base + 0x3c2) = 0x03;

//not needed?
//set attr ctrl palette reg0 + disable DAC access to palette RAM
//(dummy read to select index register)
*(dev->pcio_base + 0x3da);
//(select indexed register 'palette register 0' and disable DAC access to palette RAM)
*(dev->pcio_base + 0x3c0) = 0x00;

//select attr ctrl reg index $20
//(dummy read to select index register)
*(dev->pcio_base + 0x3da);
//(select indexed register)
*(dev->pcio_base + 0x3c0) = 0x20;
//(dummy read to re-select index register)
*(dev->pcio_base + 0x3da);

			//turn sync, screen and logic on:
			//(REPAINT1: hsync + vsync ON, disable large screen (>=1280),
			// palette width = 8 bits)
			*(dev->pcio_base + AdresReg) = 0x1a;
			*(dev->pcio_base + DataReg) = 0x04;
			//(SEQ: clocking mode: turn screen and logic on, hor CRTC counts 8 bits,
			// VGA 'hi-res' mode)			
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) = 0x00;
			//VGA reg $21... (enable reprogram)
			*(dev->pcio_base + AdresReg) = 0x21;
			*(dev->pcio_base + DataReg) = 0xfa;

//not needed?
//clear scratch1 register)
*(dev->pcio_base + AdresReg) = 0x2c;
*(dev->pcio_base + DataReg) = 0x00;
			
			//setup SEQUENCER registers:
			//(place VGA chip in 'reset state' during SEQ reprogram)
			*(dev->pcio_base + 0x3c4) = 0x00;
			*(dev->pcio_base + 0x3c5) = 0x01;
			//(non-seq) misc. output: (set already to color mode earlier)
			*(dev->pcio_base + 0x3c2) = 0x03;//sync pol = not used, xtal src = not used
			//SEQ: clocking mode setup:
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) = 0x00;//no responce for 01 or 0x09?
			//SEQ: map mask setup:
			*(dev->pcio_base + 0x3c4) = 0x02;
			*(dev->pcio_base + 0x3c5) = 0x00;//not used: old VGA plane access
			//SEQ: character map setup:
			*(dev->pcio_base + 0x3c4) = 0x03;
			*(dev->pcio_base + 0x3c5) = 0x00;//not used: text-mode VGA
			//SEQ: memory mode setup:
			*(dev->pcio_base + 0x3c4) = 0x04;
			*(dev->pcio_base + 0x3c5) = 0x0e;//seq. adressing, >64kRAM, graph. mode
			//(end VGA chip 'reset state')
			*(dev->pcio_base + 0x3c4) = 0x00;
			*(dev->pcio_base + 0x3c5) = 0x03;

//not needed?
//$3d4/5 reg at index $11 b5 = 1 = non-standard!!:
*(dev->pcio_base + AdresReg) = 0x11;
*(dev->pcio_base + DataReg) |= 0x20;

//crt ctrl regs 25 setup:
//already happens below..

//std graphics ctrl regs (9) setup:
//already happens below.. (all resetted there)

			//(SEQ: clocking mode: turn screen and logic OFF
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) |= 0x20;

//not needed?
//misc output setup again $3c2

//seq index $01 clock mode again

	//unlock VGA registers at index 0-7
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) &= 0x7f;

			//setup std VGA CTRC registers:
			//(VESA 800x600 mode for now..)
			//hor. display end: (nr. displayed pixels hor)
			*(dev->pcio_base + AdresReg) = 0x01;
			*(dev->pcio_base + DataReg) = 0x63;
			//max. scanline reg:
			//(set vertical res. to normal operation (prevent /2 = vert scale 200%))
			*(dev->pcio_base + AdresReg) = 0x09;
			*(dev->pcio_base + DataReg) = 0x00;//no influence because overwritten later!

			//clear non-programmed bits in LSR
			*(dev->pcio_base + AdresReg) = 0x25;
			*(dev->pcio_base + DataReg) = 0x00;
			//clear non-programmed bits in HEB
			*(dev->pcio_base + AdresReg) = 0x2d;
			*(dev->pcio_base + DataReg) = 0x00;

	//copy of VESA VGA 800x600 mode:			
				//800x600 resolution active...
				lsb = 0x72;//values for vertical total...
				msb = 0x02;

				//values for RIVA pixelclock (60Hz):
				temp32 = 0x00027b0b;

				//set hor blank start:
				*(dev->pcio_base + AdresReg) = 0x02;
				*(dev->pcio_base + DataReg) = 0x63;

				//set hor restrace start:
				*(dev->pcio_base + AdresReg) = 0x04;
				*(dev->pcio_base + DataReg) = 0x6a;

				//set hor display total:
				*(dev->pcio_base + AdresReg) = 0x00;
				*(dev->pcio_base + DataReg) = 0x7f;

				//set hor retrace end:
				*(dev->pcio_base + AdresReg) = 0x05;
				*(dev->pcio_base + DataReg) = 0x1a;

				//set hor blank end:
				*(dev->pcio_base + AdresReg) = 0x03;
				*(dev->pcio_base + DataReg) = 0x83;

				//set LSR:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vertical display end:
				*(dev->pcio_base + AdresReg) = 0x12;
				*(dev->pcio_base + DataReg) = 0x57;

				//set vert blank start:
				*(dev->pcio_base + AdresReg) = 0x15;
				*(dev->pcio_base + DataReg) = 0x57;

				//set overflow:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0xf0;

				//set height:
				*(dev->pcio_base + AdresReg) = 0x09;
				*(dev->pcio_base + DataReg) = 0x60;//b7=1 = double scan werkt perfect!

				//unlock VGA vertical retrace start and end registers for PCI access!
				//(nessesary here per se for PCI access:
				// access expires on writing to other registers than just those two...)
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) &= 0x7f;
				*(dev->pcio_base + AdresReg) = 0x1f;
				*(dev->pcio_base + DataReg) = 0x57;

				//unlock VGA registers 'vert. retrace start' & 'vert. retrace end' for ISA access:
				//(may be placed 'everywhere' up to this point in the code:
				// access stays enabled until locked manually again...)
//				ISA_Bus->write_io_8(AdresReg,0x03);
//				ISA_Bus->write_io_8(DataReg,
//					(ISA_Bus->read_io_8(DataReg) | 0x80));

				//set vert retrace start:
				*(dev->pcio_base + AdresReg) = 0x10;
				*(dev->pcio_base + DataReg) = 0x59;

				//set overflow vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x07;
				*(dev->pcio_base + DataReg) = 0xf0;

				//set LSR vertical retrace start part:
				*(dev->pcio_base + AdresReg) = 0x25;
				*(dev->pcio_base + DataReg) = 0x00;

				//set vert. retrace end:
				*(dev->pcio_base + AdresReg) = 0x11;
				*(dev->pcio_base + DataReg) = 0x0d;

				//set vert. blank end:
				*(dev->pcio_base + AdresReg) = 0x16;
				*(dev->pcio_base + DataReg) = 0x73;
						
			//Set the vertical total:
    		//LSB (vertical total):
			*(dev->pcio_base + AdresReg) = 0x06;
			*(dev->pcio_base + DataReg) = lsb;
			//b8 and b9:(overflow)
			*(dev->pcio_base + AdresReg) = 0x07;
			*(dev->pcio_base + DataReg) = 
				((*(dev->pcio_base + DataReg) & ~0x21) | ((msb & 0x01)+((msb & 0x02)<<4)));
			//b10:(LSR)
			*(dev->pcio_base + AdresReg) = 0x25;
			*(dev->pcio_base + DataReg) = 
				((*(dev->pcio_base + DataReg) & ~0x01) | ((msb & 0x04)>>2));

			//set RIVA pixelclock:
			//(MEMADR $00680508)
			//warning: the value here is dependent on the mounted XTAL on the RIVA!!
			//So: needs correction...
			*(dev->regs + (0x00680508 >> 2)) = temp32;
	//end copy VESA VGA 800x600 mode.

			//reset all graphics controller registers for straight std. VGA mode.
			for (index = 0; index < 9; index++)
			{
				*(dev->pcio_base + 0x3ce) = index;
				*(dev->pcio_base + 0x3cf) = 0x00;
			}

			//

//rudtest
//misc registers that are still missed?!?...
	*(dev->pcio_base + AdresReg) = 0x17;
	*(dev->pcio_base + DataReg) = 0xe3;//VESA val: display width granularity etc.

//three registers with the graphics/alpha bit that is redundant...
	*(dev->pcio_base + 0x3da);//mode ctrl
	*(dev->pcio_base + 0x3c0) = 0x10;
	*(dev->pcio_base + 0x3c0) = 0x01;//graphics mode
	
	*(dev->pcio_base + 0x3ce) = 0x06;//misc
	*(dev->pcio_base + 0x3cf) = 0x01;//graphics mode

	//(SEQ: graph. mode on EGA
	*(dev->pcio_base + 0x3c4) = 0x04;
	*(dev->pcio_base + 0x3c5) &= ~0x01;//graphics mode
//end rudtest


			//offset reg: specifies width of display!!!
			*(dev->pcio_base + AdresReg) = 0x13;
			*(dev->pcio_base + DataReg) = 0xc8;//800 bits: (0x63 +1)*2 !!
						
			//set framebuffer start adress and offset:
			//repaint0: reset (buffer start adress is set later on..)
			*(dev->pcio_base + AdresReg) = 0x19;
			*(dev->pcio_base + DataReg) = 0x00;
			//(repaint1: already done before.)

			//fifo control setup: burst length = 128
			//note: the optimal value depends on:
			//-color depth
			//-ram-bus width on VGA card (64 or 128 bit)
			*(dev->pcio_base + AdresReg) = 0x1b;
			*(dev->pcio_base + DataReg) = 0x03;
			
			//enh. register: ?
			*(dev->pcio_base + AdresReg) = 0x1c;
			*(dev->pcio_base + DataReg) = 0x00; //b3 and 4: no infl. (rest should be '0')

			//page0: reset
			*(dev->pcio_base + AdresReg) = 0x1d;
			*(dev->pcio_base + DataReg) = 0x00;
			//page1: reset
			*(dev->pcio_base + AdresReg) = 0x1e;
			*(dev->pcio_base + DataReg) = 0x00;

			//FFLWM register: setup watermark
			//probably also dependend on color depth and RAM-width (see fifo ctrl)
			*(dev->pcio_base + AdresReg) = 0x20;
			*(dev->pcio_base + DataReg) = 0x20;
			
			//setup PIXEL: (colorspace = 32bit)
			*(dev->pcio_base + AdresReg) = 0x28;
			*(dev->pcio_base + DataReg) = 0x02;//0x03 = 32 bit<<<
			
			//page-ovfl: reset
			*(dev->pcio_base + AdresReg) = 0x29;
			*(dev->pcio_base + DataReg) = 0x00;
			//oscol: reset
			*(dev->pcio_base + AdresReg) = 0x2a;
			*(dev->pcio_base + DataReg) = 0x00;

			//hcur-addr0: reset
			*(dev->pcio_base + AdresReg) = 0x30;
			*(dev->pcio_base + DataReg) = 0x00;
			//hcur-addr1: reset
			*(dev->pcio_base + AdresReg) = 0x31;
			*(dev->pcio_base + DataReg) = 0x00;
			//vid-end0: reset
			*(dev->pcio_base + AdresReg) = 0x32;
			*(dev->pcio_base + DataReg) = 0x00;

			//lcd setup: no stereo, no lcd output
			*(dev->pcio_base + AdresReg) = 0x33;
			*(dev->pcio_base + DataReg) = 0x00;

			//interlace setup
			*(dev->pcio_base + AdresReg) = 0x39;
			*(dev->pcio_base + DataReg) = 0xff;//VESA does this also..
			//?? b6 en 7 doen iets: l<>r omgedraaid!? framebuffer start adress?
			//0x00 = 120hz, interlaced / 0xc0 = 60Hz, non-interlaced

			//switch RIVA onto VGA://warning: waits a few images!!
			LEAVE_CRITICAL_SECTION
			SwitchToVGA(dev);
			ENTER_CRITICAL_SECTION

			//PRAMDAC general control setup:
			//-select VGA state
			//-75 Ohm termination
			//-alt mode 16
			//- BPC 8 bit
			*(dev->regs + (0x00680600 >> 2)) = 0x00101100;//VESA setting.

			//PFB-config-0:
			//-tiling disabled
			*(dev->regs + (0x00100200 >> 2)) = 0x00001000;

//select index $18 lcomp (does nothing?):
*(dev->pcio_base + AdresReg) = 0x18;

			//FP_TG_CTRL:
			//-control mode = native
			temp32 = *(dev->regs + (0x00680848 >> 2)) &= ~0x00000100;
			*(dev->regs + (0x00680848 >> 2)) = (temp32 | 0x00000200);

	//copy of setFBout:
				//create fbw at RAM offset 0x0:
				data->fb.fbw = (vuint32*) 0;
				//the fbw framebuffer is currently displayed:
				data->fb.fbout_used = 2;

				//schedule original windowmanager framebuffer:
				//b1-0:
				//(select VGA ATTR index register:)
				*(dev->pcio_base + 0x3da);
				//(select ATTR register to write:)
				*(dev->pcio_base + 0x3c0) = 0x33;
				//(write data:)
				*(dev->pcio_base + 0x3c0) = ((((uint32)data->fb.fbw) << 1) & 0x06);
				//b17-10 goto SA_HI:
				*(dev->pcio_base + AdresReg) = 0x0c;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fbw) >> 10);
				//b2-9 goto SA_LO:
				*(dev->pcio_base + AdresReg) = 0x0d;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fbw) >> 2);
				//b22-18 goto repaint0:
				*(dev->pcio_base + AdresReg) = 0x19;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0xe0) |
					 ((((uint32)data->fb.fbw) >> 18) & 0x1f));
				//b24-23 goto HEB:
				*(dev->pcio_base + AdresReg) = 0x2d;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0x9f) |
					 ((((uint32)data->fb.fbw) >> 18) & 0x60));
	//end copy setFBout.

			//(SEQ: clocking mode: turn screen and logic on, hor CRTC counts 8 bits,
			// VGA 'hi-res' mode)			
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) = 0x01;//09: does not work!!

			//SEQ: clocking mode: turn screen and logic ON			
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) &= ~0x20;

//select attr ctrl reg index $20
//(dummy read to select index register)
*(dev->pcio_base + 0x3da);
//(select indexed register)
*(dev->pcio_base + 0x3c0) = 0x20;
//(dummy read to re-select index register)
*(dev->pcio_base + 0x3da);

			//scratch1: select flat framebuffer access mode			
			*(dev->pcio_base + AdresReg) = 0x2c;
			*(dev->pcio_base + DataReg) |= 0x40;

	//place some garbage in the framebuffer (test):
			//hor. 'white' line 400 pixels wide, hor. centered:
			for (index32 = 0; index32 < 800; index32++)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*150 + 400 + index32)
					= 0xff;
			}
			//left 'blue' vertical line:
			for (index32 = 0; index32 < 300; index32++)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*150 + 400 + index32*1600)
					= 0xff;
			}
			//hor. white line 400 pixels wide, hor. centered:
			for (index32 = 0; index32 < 800; index32++)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*450 + 400 + index32)
					= 0xff;
			}
			//right 'blue' vertical line:
			for (index32 = 0; index32 < 300; index32++)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*150 + 1200 + index32*1600)
					= 0xff;
			}
			//centered 'yellow' vertical line:
			for (index32 = 0; index32 < 599; index32++)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*0 + 801 + index32*1600)
					= 0xff;
			}
			//hor. 'yellow' line 400 pixels wide, hor. centered:
			for (index32 = 1; index32 < 1600; index32 +=2)
			{
				*(vuint32*)((uint32)data->fb.fbw + (uint32)data->fb.fbmout
				 + 1600*300 + 0 + index32)
					= 0xff;
			}
	//end of the garbage image.
	
			LEAVE_CRITICAL_SECTION
			break;
	case rivaSetFBOut:
    		dprintf("riva_iic: setting new framebuffer.\n");  
			ENTER_CRITICAL_SECTION
			if (data->fb.fbout_used == 2)
			{
				//schedule original windowmanager framebuffer:
				//b1-0:
				//(select VGA ATTR index register:)
				*(dev->pcio_base + 0x3da);
				//(select ATTR register to write:)
				*(dev->pcio_base + 0x3c0) = 0x33;
				//(write data:)
				*(dev->pcio_base + 0x3c0) = ((((uint32)data->fb.fbw) << 1) & 0x06);
				//b17-10 goto SA_HI:
				*(dev->pcio_base + AdresReg) = 0x0c;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fbw) >> 10);
				//b2-9 goto SA_LO:
				*(dev->pcio_base + AdresReg) = 0x0d;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fbw) >> 2);
				//b22-18 goto repaint0:
				*(dev->pcio_base + AdresReg) = 0x19;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0xe0) |
					 ((((uint32)data->fb.fbw) >> 18) & 0x1f));
				//b24-23 goto HEB:
				*(dev->pcio_base + AdresReg) = 0x2d;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0x9f) |
					 ((((uint32)data->fb.fbw) >> 18) & 0x60));
			}
			else if (data->fb.fbout_used == 1)
			{
				//schedule framebuffer 2:
				//b1-0:
				//(select VGA ATTR index register:)
				*(dev->pcio_base + 0x3da);
				//(select ATTR register to write:)
				*(dev->pcio_base + 0x3c0) = 0x33;
				//(write data:)
				*(dev->pcio_base + 0x3c0) = ((((uint32)data->fb.fb2out) << 1) & 0x06);
				//b17-10 goto SA_HI:
				*(dev->pcio_base + AdresReg) = 0x0c;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fb2out) >> 10);
				//b2-9 goto SA_LO:
				*(dev->pcio_base + AdresReg) = 0x0d;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fb2out) >> 2);
				//b22-18 goto repaint0:
				*(dev->pcio_base + AdresReg) = 0x19;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0xe0) |
					 ((((uint32)data->fb.fb2out) >> 18) & 0x1f));
				//b24-23 goto HEB:
				*(dev->pcio_base + AdresReg) = 0x2d;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0x9f) |
					 ((((uint32)data->fb.fb2out) >> 18) & 0x60));
			}
			else
			{
				//schedule framebuffer 1:
				//b1-0:
				//(select VGA ATTR index register:)
				*(dev->pcio_base + 0x3da);
				//(select ATTR register to write:)
				*(dev->pcio_base + 0x3c0) = 0x33;
				//(write data:)
				*(dev->pcio_base + 0x3c0) = ((((uint32)data->fb.fb1out) << 1) & 0x06);
				//b17-10 goto SA_HI:
				*(dev->pcio_base + AdresReg) = 0x0c;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fb1out) >> 10);
				//b2-9 goto SA_LO:
				*(dev->pcio_base + AdresReg) = 0x0d;
				*(dev->pcio_base + DataReg) = (((uint32)data->fb.fb1out) >> 2);
				//b22-18 goto repaint0:
				*(dev->pcio_base + AdresReg) = 0x19;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0xe0) |
					 ((((uint32)data->fb.fb1out) >> 18) & 0x1f));
				//b24-23 goto HEB:
				*(dev->pcio_base + AdresReg) = 0x2d;
				*(dev->pcio_base + DataReg) =
					((*(dev->pcio_base + DataReg) & 0x9f) |
					 ((((uint32)data->fb.fb1out) >> 18) & 0x60));
			}
			//issue actual schedule:
			*(dev->pcio_base + AdresReg) = 0x11;
			*(dev->pcio_base + DataReg) &= 0xef;
 			LEAVE_CRITICAL_SECTION
			break;
	case rivaChkFBOutToggle:
    		dprintf("riva_iic: checking framebuffer toggle.\n");  
			ENTER_CRITICAL_SECTION
			data->fb.fbout_sw = ((*(dev->pcio_base + 0x03c2) & 0x80) >> 7);
 			LEAVE_CRITICAL_SECTION
			break;
	case rivaColdStart:
			ENTER_CRITICAL_SECTION
	//coldstart pre-init:
	  		dprintf("riva_iic: Attempting card coldstart...\n");  

//extra test: PCI config space: try to force ISA access on..
//NV_PBUS_PCI_NV_1
*(dev->regs + (0x00001804 >> 2)) |= 0x00000007;


			//setup VIO_VSE2 register:
			*(dev->pcio_base + 0x3c3) = 0x01;

			//set card to color mode (sets registers to expected adresses)
			//misc. output:
			//(set color mode and enable RAM access)
			*(dev->pcio_base + 0x3c2) = 0x03;

//extra test: unlock VGA extended registers...
*(dev->pcio_base + AdresReg) = 0x1f;
*(dev->pcio_base + DataReg) = 0x57;

			//turn off RMA:
			*(dev->pcio_base + AdresReg) = 0x38;
			*(dev->pcio_base + DataReg) = 0x00;

//extra test: read reg to determine secam/non-secam mode setup required
//P_EXTDEV_BOOT0:
*(dev->regs + (0x00101000 >> 2));

			//set scratch0 for non-secam (standard) mode
			*(dev->pcio_base + AdresReg) = 0x2b;
			*(dev->pcio_base + DataReg) = 0x00;

//IIC pre-init all connected chips.
//setup TV black output.

//set std VGA mode:
*(dev->pcio_base + AdresReg) = 0x28;
*(dev->pcio_base + DataReg) &= 0x7b;
//enable reprogram:
*(dev->pcio_base + AdresReg) = 0x21;
*(dev->pcio_base + DataReg) &= 0xfa;

//switch RIVA pixelclock to be its own:
			//NV_PRAMDAC_TV_SETUP
			//set: TV_SETUP_DEV_TYPE_SLAVE
			*(dev->regs + (0x00680700 >> 2)) &= ~0x00000001;
			//NV_PRAMDAC_PLL_COEFF_SELECT:
			//SELECT_VS_PCLK_TV_NONE;
			//SELECT_TVCLK_SOURCE_EXT;
			//SELECT_TVCLK_RATIO_DB1.
			*(dev->regs + (0x0068050c >> 2)) &= 0xfeecffff;
			//TREG:
			*(dev->pcio_base + AdresReg) = 0x3d;
			*(dev->pcio_base + DataReg) = 0x00;
//end switch pixelclock.			

//card specific actual coldstart:
//WARNING: this part is probably not valid for GeForce cards!
//(Values correspond to settings in ELSA Erazor III TNT2 card.)

			//1.
			//remember cards settings PCI config space #19:
			temp32 = *(dev->regs + (0x0000184C >> 2));
			//turn card off via PCI config space #19:
			*(dev->regs + (0x0000184C >> 2)) = 0x00000000;
			//PMC_ENABLE: disable PFB
			*(dev->regs + (0x00000200 >> 2)) = 0x01010100;
			//PMC_ENABLE: enable PFB
			*(dev->regs + (0x00000200 >> 2)) = 0x01110100;
			//turn card back to original PCI config space #19 settings:
			*(dev->regs + (0x0000184C >> 2)) = temp32;

			//2.
			//PEXTDEV_BOOT0 modify:
			//b11 = AGP_FASTWR_DISABLED
			//b31 = STRAP_OVERWRITE_ENABLED
			*(dev->regs + (0x00101000 >> 2)) |= 0x80000800;
			
			//3.
			//PMC_INTR0: clear all INT flags:
			*(dev->regs + (0x00000100 >> 2)) = 0x00000000;

			//4.
			//PBUS_DEBUG_PRIV_ASRC:
			*(dev->regs + (0x000010e0 >> 2)) = 0x00000000;

			//5.
			//PBUS_DEBUG0:
			//b16-19: FBIO_ACLK_DELAY = $e
			//b24-27: FBIO_RCLK_DELAY = $4
			*(dev->regs + (0x00001080 >> 2)) = 0x040e0000;
			
			//6.
			//PBUS_DEBUG_CTRIM0:
			*(dev->regs + (0x000010b0 >> 2)) = 0x0800bba7;

			//7.
			//PBUS_DEBUG_CTRIM1:
			*(dev->regs + (0x000010b4 >> 2)) = 0xefe000ff;

			//8.
			//PBUS_DEBUG_CTRIM2:
			*(dev->regs + (0x000010b8 >> 2)) = 0x88c98200;

			//9.
			//PBUS_DEBUG_CTRIM3:
			*(dev->regs + (0x000010bc >> 2)) = 0xff988aa9;

			//10.
			//NVPLL_COEFF setup:
			*(dev->regs + (0x00680500 >> 2)) = 0x0001e30d;

			//11.
			//MPLL_COEFF setup:
			*(dev->regs + (0x00680504 >> 2)) = 0x00005809;

			//12.
			//PFB_CONFIG0:
			//type = no tiling
			*(dev->regs + (0x00100200 >> 2)) = 0x00001114;

			//13.
			//VPLL_COEFF setup:
			*(dev->regs + (0x00680508 >> 2)) = 0x0003c50e;

			//14.
			//unknown reg clear:
			*(dev->regs + (0x001000c0 >> 2)) = 0x00000000;

			//15.
			//PFB_DEBUG0:
			//b8-13: refresh count *64 = $25 (std = $10)
			//b29: refinc = enabled
			*(dev->regs + (0x00100080 >> 2)) = 0x20002500;

			//16.
			//PFB_CONFIG1:
			//setup RAM timing (RAS/CAS/delays...)
			*(dev->regs + (0x00100204 >> 2)) = 0x41016293;
			
			//17.
			//PFB_RTL:
			*(dev->regs + (0x00100300 >> 2)) = 0x00000000;

			//18.
			//PBUS_DEBUG1:
			//setup RD/WR BURST, decode NV4...)
			*(dev->regs + (0x00001084 >> 2)) = 0x000057c8;

			//19.
			//RAM size and width test, including presetup via table.
			testRAM(dev);

			//20.
			//PFB_BOOT0:
			//does not modify data
			*(dev->regs + (0x00100000 >> 2)) |= 0x00000000;

			//21.
			//PFB_CONFIG1:
			*(dev->regs + (0x00100204 >> 2)) = 0xc1015283;

			//22.
			//PRAMDAC_PLL_COEFF_SELECT:
			//b8-10: source = PROG_MPLL + PROG_NVPLL
			*(dev->regs + (0x0068050c >> 2)) = 0x00000500;

			//23.
			//PBUS_DEBUG2:
			//b4-9: AGP_SB_STB_DELAY = $1f
			*(dev->regs + (0x00001088 >> 2)) = 0x000001f0;

			//24.
			//PBUS_DEBUG3:
			//b0-1: AGP_MAX_SIZE = 32 bytes
			//b4-7: AGP_4x_NV_CLK = above 120MHz
			*(dev->regs + (0x0000108c >> 2)) = 0x000000d1;

	//not needed?
			//25.
			//RMA reset:
			*(dev->pcio_base + AdresReg) = 0x38;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0x00;

			//26.
			//REPAINT1 setup:
			//- enable hsync and vsync
			*(dev->pcio_base + AdresReg) = 0x1a;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0x3f;

			//27.
			//PIXEL:
			//- std VGA
			*(dev->pcio_base + AdresReg) = 0x28;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0x00;

			//28.
			//LCD reset:
			//- disable LCD
			//- disable stereoscopic view
			*(dev->pcio_base + AdresReg) = 0x33;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0x00;

			//29.
			//enable reprogram:
			*(dev->pcio_base + AdresReg) = 0x21;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0xfa;

			//30.
			//SCRATCH0:
			//- clear unused bits
			*(dev->pcio_base + AdresReg) = 0x2b;
			//read + write
			*(dev->pcio_base + DataReg) &= 0x07;

			//31.
			//SCRATCH1:
			*(dev->pcio_base + AdresReg) = 0x2c;
			//read
			*(dev->pcio_base + DataReg);
			//write
			*(dev->pcio_base + DataReg) = 0x00;
	//end not needed?

			//setup PRAMDAC_FP_TG_CTRL reg:
			//mode = native (or centered or scaled ?)
			//read prog
			//dispen = pos (or neg ?)
			temp32 = 0x10100200;
			//check interface for 12/24bit via P_EXTDEV_BOOT0, init accordingly:
			//(only needed for using devices connected to RIVA pixeldata port)
			if (*(dev->regs + (0x00101000 >> 2)) & 0x00008000)
				temp32 |= 0x01000000;
			*(dev->regs + (0x00680848 >> 2)) = temp32;

			//setup PRAMDAC_FP_DEBUG0 reg:
			//pwrdwn = FPCLK
			//(pwrdwn = FPCLK means TV out not used:
			*(dev->regs + (0x00680880 >> 2)) |= 0x10000000;
	//end coldstart pre-init. 

//note: VGA termination should be set to 37.5 Ohm if no VGA monitor is detected.
//this should be checked via the VESA DDC channel.
//(correct termination prevents 'standing wave' problems (overheating) of the RIVA chip)
//PRAMDAC GENERAL CONTROL reg, b17 controls the termination. (0=37, 1=75 Ohm)

//next to do after this routine:
//-setup a mode
//-setup HW cursor if needed
//-switch to TV or VGA.

 			LEAVE_CRITICAL_SECTION
			break;
    default:dprintf("riva_iic: illegal ioctrl command\n");
            err = B_DEV_INVALID_IOCTL;
            break;
    }        
    
	dprintf("leave iic_device_control\n");
	return(err);
}


static status_t
iic_device_close (void* cookie)
{
	rivapci_private_t *dev = (rivapci_private_t *) cookie;

	/* unmap card from memory */
	delete_area(dev->ioarea);
	dev->ioarea = -1;
	delete_area(dev->fbarea);
	dev->fbarea = -1;

	dprintf("riva_iic: device closed\n");
	return B_OK;
}


static status_t
iic_device_free (void* cookie)
{
	return B_OK;
}


static const char *my_device_name[] = 
{
	"video/riva_iic/1",
	NULL
};


device_hooks my_device_hooks = 
{
	iic_device_open, 			/* -> open entry point */
	iic_device_close, 			/* -> close entry point */
	iic_device_free,			/* -> free cookie */
	iic_device_control, 		/* -> control entry point */
	iic_device_read,			/* -> read entry point */
	iic_device_write			/* -> write entry point */
};


const char**
publish_devices()
{
	return my_device_name;
}


device_hooks*
find_device(const char* name)
{
	return &my_device_hooks;
}

void RMA_PCI_Write(rivapci_private_t *dev, uint32 adress, uint32 data)
{
	//write adress:
	//RMA:
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x03;
	//MEMADR:
/*
	*(dev->pcio_base + 0x03d0) = ((uint8)adress);
	*(dev->pcio_base + 0x03d1) = ((uint8)(adress >> 8));
	*(dev->pcio_base + 0x03d2) = ((uint8)(adress >> 16));
	*(dev->pcio_base + 0x03d3) = ((uint8)(adress >> 24));
*/
	//does not work via PCI access, so use ISA access for now:
    ISA_Bus->write_io_8(0x03d0,((uint8)adress));
    ISA_Bus->write_io_8(0x03d1,((uint8)(adress >> 8)));
    ISA_Bus->write_io_8(0x03d2,((uint8)(adress >> 16)));
    ISA_Bus->write_io_8(0x03d3,((uint8)(adress >> 24)));
	//write data:
	//RMA:
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x07;
	//MEMADR:
/*
	*(dev->pcio_base + 0x03d0) = ((uint8)data);
	*(dev->pcio_base + 0x03d1) = ((uint8)(data >> 8));
	*(dev->pcio_base + 0x03d2) = ((uint8)(data >> 16));
	*(dev->pcio_base + 0x03d3) = ((uint8)(data >> 24));
*/
	//does not work via PCI access, so use ISA access for now:
    ISA_Bus->write_io_8(0x03d0,((uint8)data));
    ISA_Bus->write_io_8(0x03d1,((uint8)(data >> 8)));
    ISA_Bus->write_io_8(0x03d2,((uint8)(data >> 16)));
    ISA_Bus->write_io_8(0x03d3,((uint8)(data >> 24)));
	//RMA:(not nessesary?)
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x03;
}//RMA_PCI_Write.

uint32 RMA_PCI_Read(rivapci_private_t *dev, uint32 adress)
{
	uint32 data = 0;
	
	//write adress:
	//RMA:
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x03;
	//MEMADR:
/*
	*(dev->pcio_base + 0x03d0) = ((uint8)adress);
	*(dev->pcio_base + 0x03d1) = ((uint8)(adress >> 8));
	*(dev->pcio_base + 0x03d2) = ((uint8)(adress >> 16));
	*(dev->pcio_base + 0x03d3) = ((uint8)(adress >> 24));
*/
	//does not work via PCI access, so use ISA access for now:
    ISA_Bus->write_io_8(0x03d0,((uint8)adress));
    ISA_Bus->write_io_8(0x03d1,((uint8)(adress >> 8)));
    ISA_Bus->write_io_8(0x03d2,((uint8)(adress >> 16)));
    ISA_Bus->write_io_8(0x03d3,((uint8)(adress >> 24)));
	//read data:
	//RMA:
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x05;
	//MEMADR:
/*
	data = *(dev->pcio_base + 0x03d0);
	data |= (((uint32) *(dev->pcio_base + 0x03d1)) << 8);
	data |= (((uint32) *(dev->pcio_base + 0x03d2)) << 16);
	data |= (((uint32) *(dev->pcio_base + 0x03d3)) << 24);
*/
	//does not work via PCI access, so use ISA access for now:
    data = ISA_Bus->read_io_8(0x03d0);
    data |= (ISA_Bus->read_io_8(0x03d1)<< 8);
    data |= (ISA_Bus->read_io_8(0x03d2)<<16);
    data |= (ISA_Bus->read_io_8(0x03d3)<<24);
	//RMA:(not nessesary?)
	*(dev->pcio_base + AdresReg) = 0x38;
	*(dev->pcio_base + DataReg) = 0x03;
	return data;
}//RMA_PCI_Read.

void testRAM(rivapci_private_t *dev)
{
	//RAM size and width test, including presetup via table.
	uint32 temp32;
	int strapram = 0;

	//warning: this table will be card specific, 
	//works probably (best case) with cards TNT1, 2 and TNT2-M64... 
	strapstruct straptable[8] = {{0x24,0x00},
								{0x28,0x00},
								{0x24,0x01},
								{0x1f,0x00},
								{0x0f,0x00},
								{0x17,0x00},
								{0x06,0x00},
								{0x00,0x00}};

	//unlock ext VGA regs
		//unlock registers:
		*(dev->pcio_base + AdresReg) = 0x1f;
		*(dev->pcio_base + DataReg) = 0x57;//write 'magic number'

			//turn screen and logic off:
			//(SEQ: clocking mode: turn screen and logic OFF
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) |= 0x20;

			//PFB_DEBUG0:
			//b4: MRS = 256
			*(dev->regs + (0x00100080 >> 2)) &= ~0x00000010;

			//PEXTDEV_BOOT0 reg:
			//- get STRAP_RAM_CFG
		//strapram = ((*(dev->regs + (0x00101000 >> 2)) & 0x0000003c) >> 2);
		//failsave: table implemented is too short:
			strapram = ((*(dev->regs + (0x00101000 >> 2)) & 0x0000001c) >> 2);
			dprintf("riva_iic: testRAM: strapram %d hardwired.\n", strapram);

			//PFB_BOOT0:
			//setup RAM amount, width, type
			temp32 = straptable[strapram].ram;
			*(dev->regs + (0x00100000 >> 2)) = 
				((*(dev->regs + (0x00100000 >> 2)) & ~0x0000003f) | temp32);

			//PFB_CONFIG1:
			//CAS latency, Write to Read default
			temp32 = (*(dev->regs + (0x00100204 >> 2)) & ~0x00700001);
			//or-in new data:
			temp32 |= (straptable[strapram].wr_delay << 20);
			//write..
			*(dev->regs + (0x00100204 >> 2)) = temp32;
			//set b0:
			*(dev->regs + (0x00100204 >> 2)) = (temp32 | 0x00000001);

	//RAM access via RMA done on purpose:
	//prevents reading or writing outside allocated space!
	//Add 0x81000000 to every adress in memory to use this access method.

	//check width of videoRAM.
	//(that is: if card is presetted to 128 bit wide at this moment.. fixme?)
		//rudtest:
		//PFB_BOOT0:
		//set RAM width = 128 bit.
		*(dev->regs + (0x00100000 >> 2)) |= 0x00000004;
		//end rudtest.

	//write testpattern to start of videoRAM:
	RMA_PCI_Write(dev, 0x81000000, 0x4e563541);//b0-31
	RMA_PCI_Write(dev, 0x81000004, 0x4e563541);//b32-63
	RMA_PCI_Write(dev, 0x81000008, 0x4e563541);//b64-95
	RMA_PCI_Write(dev, 0x8100000c, 0x4e563541);//b96-127

	//check last written adres for correct pattern:
	//(Check fails on 64 bit wide RAM if RIVA presetted to 128 bit..)
	if (RMA_PCI_Read(dev, 0x8100000c)!= 0x4e563541)
	{
		//PFB_BOOT0:
		//set RAM width = 64 bit.
		*(dev->regs + (0x00100000 >> 2)) &= ~0x00000004;
		dprintf("riva_iic: testRAM: RAM is 64 bits wide.\n");
	}
	else
	{
		dprintf("riva_iic: testRAM: RAM is 128 bits wide.\n");
	}

	//check size of videoRAM:
		//PFB_BOOT0:
		//read preset amount of RAM.
		temp32 = (*(dev->regs + (0x00100000 >> 2)) & 0x00000003);
		
		if (!temp32)
		{
			//(32Mb RAM was preset)

			//test for > 16Mb RAM:
			//write testpattern to first adress above 16Mb:
			RMA_PCI_Write(dev, 0x82000000, 0x4e563542);
			//check if succeeded:
			if (RMA_PCI_Read(dev, 0x82000000) != 0x4e563542)
			{
				//fail: set card to 16Mb and exit.
				//PFB_BOOT0:
				*(dev->regs + (0x00100000 >> 2)) |= 0x00000003;
				dprintf("riva_iic: testRAM: setting 16Mb RAM connected.\n");
				goto ramtestDone;
			}
			else
			{
				//write 2nd testpattern to first RAM adress (foldback test?):
				RMA_PCI_Write(dev, 0x81000000, 0x4135564e);

				//check if succeeded:
				if (RMA_PCI_Read(dev, 0x81000000) != 0x4135564e)
				{
					//fail: set card to 16Mb and exit.
					//PFB_BOOT0:
					*(dev->regs + (0x00100000 >> 2)) |= 0x00000003;
					dprintf("riva_iic: testRAM: setting 16Mb RAM connected.\n");
					goto ramtestDone;
				}
				else
				{
					//ok: 32Mb detected, just exit (preset was OK).
					dprintf("riva_iic: testRAM: 32Mb RAM connected.\n");
					goto ramtestDone;
				}
			}//32Mb foldback test.
		}//32Mb test.

		if (temp32 == 3)
		{
			//(16Mb RAM was preset)
			
			//test for > 8Mb RAM:
			//write testpattern to first adress above 8Mb:
			RMA_PCI_Write(dev, 0x81800000, 0x4e563543);
			//check if succeeded:
			if (RMA_PCI_Read(dev, 0x81800000) == 0x4e563543)
			{
				//ok: 16Mb detected, just exit (preset was OK).
				dprintf("riva_iic: testRAM: 16Mb RAM connected.\n");
				goto ramtestDone;
			}
			else
			{
				//fail: preset card to 8Mb and continue with 8Mb RAM test.
				//PFB_BOOT0:
				temp32 = (*(dev->regs + (0x00100000 >> 2)) & ~0x00000003);
				*(dev->regs + (0x00100000 >> 2)) = (temp32 | 0x00000002);
			}
		}

		//(8Mb or 4Mb RAM was preset)
		
		//warning: org uses:
		//-'16Mb pattern' if we didn't get here via 16Mb check;
		//-'16Mb pattern + 1' if we did get here via 16Mb check.
		//(probably not important though as long as the patterns are different
		// for each test..)

		//test for > 4Mb RAM:
		//write testpattern to first adress above 4Mb:
		RMA_PCI_Write(dev, 0x81400000, 0x4e563544);
		//check if succeeded:
		if (RMA_PCI_Read(dev, 0x81400000) == 0x4e563544)
		{
			//ok: 8Mb detected, just exit (preset was OK).
			dprintf("riva_iic: testRAM: 8Mb RAM connected.\n");
			goto ramtestDone;
		}
		else
		{
			//fail: card has to have 4Mb RAM, so set card to 4Mb and exit.
			//PFB_BOOT0:
			temp32 = (*(dev->regs + (0x00100000 >> 2)) & ~0x00000003);
			*(dev->regs + (0x00100000 >> 2)) = (temp32 | 0x00000001);
			dprintf("riva_iic: testRAM: 4Mb RAM connected.\n");
			goto ramtestDone;
		}
		
ramtestDone:
			//turn sync, screen and logic on:
			//SEQ: clocking mode
			*(dev->pcio_base + 0x3c4) = 0x01;
			*(dev->pcio_base + 0x3c5) &= ~0x20;
}//testRAM.

void SwitchToVGA(rivapci_private_t *dev)
{
	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869)
	//and Diamond Viper V550 16Mb PCI (TNT1/BT869)

	uint32 temp32;
	unsigned char VertRetrace;
	
	ENTER_CRITICAL_SECTION
	//REPAINT1: switch on VGA monitor HSYNC and VSYNC
	*(dev->pcio_base + AdresReg) = 0x1a;
	*(dev->pcio_base + DataReg) &= 0x3f;

	//SEQ:
	*(dev->pcio_base + 0x3c4) = 0x01;
	*(dev->pcio_base + 0x3c5) = 0x01;

	//wait for 1 image to be generated to make sure VGA has kicked in and is
	//running OK before continuing...
	//(Kick in will fail often if we do not wait here: You'll notice this most
	// prominently in VESA 640x480 mode. Just try switching about 10 times and
	// you'll probably see... (re-checked for PCI V0.20))
	VertRetrace = 1;
	//(make sure we are 'in' active VGA picture:)
	do  VertRetrace = (*(dev->pcio_base + 0x3da) & 0x08);
		while (VertRetrace);
	//(wait for vertical retrace start on VGA:)
	do  VertRetrace = (*(dev->pcio_base + 0x3da) & 0x08);
		while (!VertRetrace);
	//(make sure we are 'in' active VGA picture again:)
	do  VertRetrace = (*(dev->pcio_base + 0x3da) & 0x08);
		while (VertRetrace);

	//'update' PIXEL/TV:
	*(dev->pcio_base + AdresReg) = 0x28;
	*(dev->pcio_base + DataReg) &= 0x03;

	//VGA reg $21... (enable reprogram)
	*(dev->pcio_base + AdresReg) = 0x21;
	*(dev->pcio_base + DataReg) = 0xfa;

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though... (tested on ISA...)
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	//SWITCH RIVA pixelclock to be RIVA's own (so: directly):
	//MEMADR $00680700:
	temp32 = (*(dev->regs + (0x00680700 >> 2)) & ~0x00000001);
	*(dev->regs + (0x00680700 >> 2)) = (temp32 | 0x01000000);

	//switch RIVA PLL phase-lock to lock to RIVA's own internal pixelclock:
	//MEMADR $0068050c:
	*(dev->regs + (0x0068050c >> 2)) = 0x10000700;

	//TREG:
	*(dev->pcio_base + AdresReg) = 0x3d;
	*(dev->pcio_base + DataReg) = 0x00;

	//make sure LCD is switched off:
	//MEMADR $00680880:
	*(dev->regs + (0x00680880 >> 2)) |= 0x10000000;

   	//LCD:
	*(dev->pcio_base + AdresReg) = 0x33;
	*(dev->pcio_base + DataReg) &= 0xfc;


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	*(dev->pcio_base + 0x3da);
	//(write index for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x11;
    //(write data for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x00;


	//unlock VGA registers at index 0-7
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) &= 0x7f;
	LEAVE_CRITICAL_SECTION
}//end SwitchToVGA.

void SwitchRIVAtoTV (char lsb, char msb, char type, rivapci_private_t *dev)
{
	unsigned char temp, temp1, temp2, temp3;

	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869);
	//Diamond Viper V550 16Mb PCI (TNT1/BT869);
	//and ASUS V7100 GeForce2 MX200 AGP/32Mb (CH7007)

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though... (tested only via ISA bus sofar..)
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	ENTER_CRITICAL_SECTION
	//SWITCH RIVA pixelclock to be 'via' TVout chip:
	//MEMADR $00680700:
	*(dev->regs + (0x00680700 >> 2)) |= 0x01000001;

	//switch RIVA PLL phase-lock to lock to pixelclock from TV-chip instead of
	//internal pixelclock:
	//(nessecary or display will 'shiver' on both TV and VGA.)
	//MEMADR $0068050c:
	*(dev->regs + (0x0068050c >> 2)) = 0x00030700;

	//VGA reg $21... (enable reprogram!)
	*(dev->pcio_base + AdresReg) = 0x21;
	*(dev->pcio_base + DataReg) = 0xfa;

	//SEQ:
	*(dev->pcio_base + 0x3c4) = 0x01;
	*(dev->pcio_base + 0x3c5) = 0x01;


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	*(dev->pcio_base + 0x3da);
	//(write index for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x11;
    //(write data for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x00;

	//unlock VGA registers at index 0-7
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) &= 0x7f;

	//execute RIVA picture timing formula:
	//read hor display end:
	*(dev->pcio_base + AdresReg) = 0x01;
	temp = *(dev->pcio_base + DataReg);

	//copy to hor blank start:
	*(dev->pcio_base + AdresReg) = 0x02;
	*(dev->pcio_base + DataReg) = temp;

	//add 1 and place in hor restrace start:
	*(dev->pcio_base + AdresReg) = 0x04;
	*(dev->pcio_base + DataReg) = (temp + 1);

	if (type == 3)//TNT2 M64...
	{
		//add 2 and copy b4-0 to hor retrace end:
		*(dev->pcio_base + AdresReg) = 0x05;
		*(dev->pcio_base + DataReg) =
			((*(dev->pcio_base + DataReg) & ~0x1f) | ((temp + 2) & 0x1f));

		//add 3 and place in hor display total:
		*(dev->pcio_base + AdresReg) = 0x00;
		*(dev->pcio_base + DataReg) = (temp + 3);

    }
    else//TNT1, TNT2, Geforce2...
	{
		//add 2 and place in hor display total:
		*(dev->pcio_base + AdresReg) = 0x00;
		*(dev->pcio_base + DataReg) = (temp + 2);

		//add 3 and copy b4-0 to hor retrace end:
		*(dev->pcio_base + AdresReg) = 0x05;
		*(dev->pcio_base + DataReg) =
			((*(dev->pcio_base + DataReg) & ~0x1f) | ((temp + 3) & 0x1f));
	}

	//add 6 and copy b4-0 to hor blank end:
	*(dev->pcio_base + AdresReg) = 0x03;
	*(dev->pcio_base + DataReg) =
		((*(dev->pcio_base + DataReg) & ~0x1f) | ((temp + 6) & 0x1f));

	//copy hor blank end b5 in hor retrace end:
	*(dev->pcio_base + AdresReg) = 0x05;
	*(dev->pcio_base + DataReg) =
		((*(dev->pcio_base + DataReg) & 0x7f) | (((temp + 6) & 0x20) << 2));

	//clear hor blank end b6 in LSR:
	*(dev->pcio_base + AdresReg) = 0x25;
	*(dev->pcio_base + DataReg) &= 0xef;

   	//'update' PIXEL/TV:
	*(dev->pcio_base + AdresReg) = 0x28;
	temp = (*(dev->pcio_base + DataReg) & 0xc7);
	*(dev->pcio_base + DataReg) = (temp | 0x80);

	//read vertical display end:
    //(LSB:)
	*(dev->pcio_base + AdresReg) = 0x12;
	temp = *(dev->pcio_base + DataReg);

	//b8 and b9:(overflow)
	*(dev->pcio_base + AdresReg) = 0x07;
	temp2 = (*(dev->pcio_base + DataReg) & 0x42);

	//b10:(LSR)
	*(dev->pcio_base + AdresReg) = 0x25;
	temp3 = (*(dev->pcio_base + DataReg) & 0x02);
	//(combine MSB:)
	temp1 = (((temp2 & 0x02)>>1) + ((temp2 & 0x40)>>5) + (temp3<<1));

	//copy to vert blank start:
	//LSB:
	*(dev->pcio_base + AdresReg) = 0x15;
	*(dev->pcio_base + DataReg) = temp;
	//b8:(overflow)
	*(dev->pcio_base + AdresReg) = 0x07;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x08) | ((temp1 & 0x01) << 3));
	//b9:(height)
	*(dev->pcio_base + AdresReg) = 0x09;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x20) | ((temp1 & 0x02) << 4));
	//b10:(LSR)
	*(dev->pcio_base + AdresReg) = 0x25;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x08) | ((temp1 & 0x04) << 1));

	//unlock VGA vertical retrace start and end registers for PCI access!
	//(nessesary here per se for PCI access:
	// access expires on writing to other registers than just those two...)
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) &= 0x7f;
	*(dev->pcio_base + AdresReg) = 0x1f;
	*(dev->pcio_base + DataReg) = 0x57;

	//unlock VGA registers 'vert. retrace start' & 'vert. retrace end' for ISA access:
	//(may be placed 'everywhere' up to this point in the code:
	// access stays enabled until locked manually again...)
//	ISA_Bus->write_io_8(AdresReg,0x03);
//	ISA_Bus->write_io_8(DataReg,
//		(ISA_Bus->read_io_8(DataReg) | 0x80));

	//add 1 and place in vert retrace start:
	temp++;
	if (temp == 0) temp1++;
	//LSB:
	*(dev->pcio_base + AdresReg) = 0x10;
	*(dev->pcio_base + DataReg) = temp;
	//b8 and b9:(overflow)
	*(dev->pcio_base + AdresReg) = 0x07;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x84) |
				(((temp1 & 0x01) << 2) + ((temp1 & 0x02) << 6)));
	//b10:(LSR)
	*(dev->pcio_base + AdresReg) = 0x25;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x04) | (temp1 & 0x04));

	//place '(vertical total + 1)' b3-0 in vert retrace end:
	//(This takes care of the 'cursor trash' on TNT1's...)
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & 0xf0) | ((lsb + 1) & 0x0f));

	//place '(vertical total + 1)' LSB in vert. blank end:
	//(This takes care of the 'cursor trash' on TNT1's...)
	*(dev->pcio_base + AdresReg) = 0x16;
	*(dev->pcio_base + DataReg) = (lsb + 1);
	//End RIVA Timing formula.

	//Set the vertical 'refresh rate':
    //LSB (vertical total):
	*(dev->pcio_base + AdresReg) = 0x06;
	*(dev->pcio_base + DataReg) = lsb;
	//b8 and b9:(overflow)
	*(dev->pcio_base + AdresReg) = 0x07;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x21) |
				((msb & 0x01) + ((msb & 0x02) << 4)));
	//b10:(LSR)
	*(dev->pcio_base + AdresReg) = 0x25;
	*(dev->pcio_base + DataReg) = 
			((*(dev->pcio_base + DataReg) & ~0x01) | ((msb & 0x04) >> 2));

  	//LCD:
	*(dev->pcio_base + AdresReg) = 0x33;
	*(dev->pcio_base + DataReg) &= 0xfe;

	//TREG:
	*(dev->pcio_base + AdresReg) = 0x3d;
	*(dev->pcio_base + DataReg) = 0x80;
	LEAVE_CRITICAL_SECTION
}

void EnableTVOut (rivapci_private_t *dev)
{
	//temporary, unless a better formula 'will be found' than already in riva_iic:
	//the formula used in riva_iic has to stay to keep the possibility open to enable
	//TVout on non-supported (qua videodriver) VGA chipsets in VESA mode.
	unsigned char temp;

	//tested on ELSA Erazor III 32Mb AGP (TNT2/BT869),
	//Diamond Viper V550 16Mb PCI (TNT1/BT869),
	//and ASUS V7100 GeForce2 MX200 AGP/32Mb (CH7007)

	//CAUTION:
	//On the TNT1, these memadresses apparantly cannot be read (sometimes)!;
	//write actions do succeed though... (tested only on ISA bus yet..)
	//($00680700 b1-23 and b25-31 apparantly are 'don't cares'...)

	ENTER_CRITICAL_SECTION
	//SWITCH RIVA pixelclock to be 'via' TVout chip:
	//MEMADR $00680700:
	*(dev->regs + (0x00680700 >> 2)) |= 0x01000001;

	//switch RIVA PLL phase-lock to lock to pixelclock from TV-chip instead of
	//internal pixelclock:
	//(nessecary or display will 'shiver' on both TV and VGA.)
	//MEMADR $0068050c:
	*(dev->regs + (0x0068050c >> 2)) = 0x00030700;

	//VGA reg $21... (enable reprogram!)
	*(dev->pcio_base + AdresReg) = 0x21;
	*(dev->pcio_base + DataReg) = 0xfa;

	//SEQ:
	*(dev->pcio_base + 0x3c4) = 0x01;
	*(dev->pcio_base + 0x3c5) = 0x01;


	//Set overscan color to 'black':
	//Disable this part if you're trying to center the output on TV,
	//you'll get blue overscan range color then. Use this as a guide-'line' ;-)
	//(select index adress register:)
	*(dev->pcio_base + 0x3da);
	//(write index for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x11;
    //(write data for 'overscan color' register:)
	*(dev->pcio_base + 0x3c0) = 0x00;


	//unlock VGA registers at index 0-7
	*(dev->pcio_base + AdresReg) = 0x11;
	*(dev->pcio_base + DataReg) &= 0x7f;

   	//'update' PIXEL/TV:
	*(dev->pcio_base + AdresReg) = 0x28;
	temp = (*(dev->pcio_base + DataReg) & 0xc7);
	*(dev->pcio_base + DataReg) = (temp | 0x80);

  	//LCD:
	*(dev->pcio_base + AdresReg) = 0x33;
	*(dev->pcio_base + DataReg) &= 0xfe;

	//TREG:
	*(dev->pcio_base + AdresReg) = 0x3d;
	*(dev->pcio_base + DataReg) = 0x80;
	LEAVE_CRITICAL_SECTION
}


//IIC implementation (bit-banger algorithm):

char FlagIICError (char ErrNo)
//error code list:
//1 - SCL locked low by device (bus is still busy)
//2 - SDA locked low by device (bus is still busy)
//3 - No Acknowledge from device (no handshake)
//4 - SDA not released for master to generate STOP bit
{
	static char IICError = 0;

	if (!IICError) IICError = ErrNo;
	if (ErrNo == -1) IICError = 0;
	return IICError;
}

void bstart (int BusNR, rivapci_private_t *dev)
{
	dprintf ("riva IICbus: START condition generated on bus %d.\n",BusNR);
	//spinlock..
	OutSDA(1,BusNR, dev); //make sure SDA is high
	spin(3);
	OutSCL(1,BusNR, dev);
	spin(3); //'wait' 3uS
	if (!InSCL(BusNR, dev)) FlagIICError (1);
	spin(6);
	//clear SDA while SCL set (bus-start condition):
	OutSDA(0,BusNR, dev);
	spin(6);
	OutSCL(0,BusNR, dev);
	spin(6);
	//clear spinlock..
}

void bstop (int BusNR, rivapci_private_t *dev)
{
	dprintf ("riva IICbus: STOP condition generated on bus %d.\n",BusNR);
	//spinlock..
	OutSDA(0,BusNR, dev); //make sure SDA is low
	spin(3);
	OutSCL(1,BusNR, dev);
	spin(3);
	if (!InSCL(BusNR, dev)) FlagIICError (1);
	spin(6);
	//set SDA while SCL set (bus-stop condition):
	OutSDA(1,BusNR, dev);
	spin(3);
	if (!InSDA(BusNR, dev)) FlagIICError (4);
	spin(3);
	//clear spinlock..
}

void TXBit (char Bit, int BusNR, rivapci_private_t *dev)
{
	//spinlock..
	//send out databit:
	if (Bit)
	{
		OutSDA(1, BusNR, dev);
		dprintf ("1");
		spin(3);
		if (!InSDA(BusNR, dev)) FlagIICError (2);
	}
	else
	{
		OutSDA(0, BusNR, dev);
		dprintf ("0");
	}
	//generate clock pulse:
	spin(6);
	OutSCL(1,BusNR, dev);
	spin(3);
	if (!InSCL(BusNR, dev)) FlagIICError (1);
	spin(6);
	OutSCL(0,BusNR, dev);
	spin(6);
	//clear spinlock..
}

char RXBit (int BusNR, rivapci_private_t *dev)
{
	char Bit = 0;
	
	//spinlock..
	OutSDA(1,BusNR, dev); //set SDA so input is possible
	//generate clock pulse:
	spin(6);
	OutSCL(1,BusNR, dev);
	spin(3);
	if (!InSCL(BusNR, dev)) FlagIICError (1);
	spin(3);
	//read in databit:
	if (InSDA(BusNR, dev))
	{
		Bit = 1;
		dprintf ("1");
	}
	else
	{
		dprintf ("0");
	}
	OutSCL(0,BusNR, dev);
	spin(6);
	//clear spinlock..
	return Bit;
}

void TXByte (char Byte, int BusNR, rivapci_private_t *dev)
//transmit one byte starting with MSbit.
//Afterwards check for low ACK signal from slave device.
{
	unsigned char BitCount, Bit;
	
	dprintf ("riva IICbus %d: TX %%",BusNR);
	for (BitCount = 8; BitCount > 0; BitCount--)
	{
		Bit = (Byte & 0x80);
		TXBit (Bit, BusNR, dev);
		Byte <<= 1;
	}
	dprintf (", RX ACK=");
	Bit = RXBit (BusNR, dev);
	if (Bit) FlagIICError (3);
	dprintf (", bus status %d.\n", FlagIICError(0));

}

char RXByte (char Ack, int BusNR, rivapci_private_t *dev)
//receive one byte starting with MSbit.
//Afterwards send (low or high) ACK signal to slave device.
{
	unsigned char BitCount, Bit;
	char Byte = 0;
	
	dprintf ("riva IICbus %d: RX %%",BusNR);
	for (BitCount = 8; BitCount > 0; BitCount--)
	{
		Byte <<= 1;
		Bit = RXBit (BusNR, dev);
		Byte += Bit;
	}
	dprintf (", TX ACK=");
	TXBit (Ack, BusNR, dev);
	dprintf (", bus status %d.\n", FlagIICError(0));
	return Byte;
}


//low-level hardware manipulation functions:

void OutSCL(char Bit, int BusNR, rivapci_private_t *dev)
{
	unsigned char Data;
	
	//spinlock...
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = WRIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = WRIICBus0;
	}

	Data = *(dev->pcio_base + DataReg) & 0xf0;
	if (Bit)
		Data |= 0x20;
	else
		Data &= ~0x20;

	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = WRIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = WRIICBus0;
	}

	*(dev->pcio_base + DataReg) = Data | 0x01;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
}

void OutSDA(char Bit, int BusNR, rivapci_private_t *dev)
{
	unsigned char Data;

	//spinlock...
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = WRIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = WRIICBus0;
	}

	Data = *(dev->pcio_base + DataReg) & 0xf0;
	if (Bit)
		Data |= 0x10;
	else
		Data &= ~0x10;

	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = WRIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = WRIICBus0;
	}

	*(dev->pcio_base + DataReg) = Data | 0x01;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
}

char InSCL(int BusNR, rivapci_private_t *dev)
{
	unsigned char Data = 0;
	
	//spinlock
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = RDIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = RDIICBus0;
	}

	if (*(dev->pcio_base + DataReg) & 0x04)
		Data = 1;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
	return Data;
}

char InSDA(int BusNR, rivapci_private_t *dev)
{
	unsigned char Data = 0;
	
	//spinlock
	ENTER_CRITICAL_SECTION
	if (BusNR)
	{
		*(dev->pcio_base + AdresReg) = RDIICBus1;
	}
	else
	{
		*(dev->pcio_base + AdresReg) = RDIICBus0;
	}

	if (*(dev->pcio_base + DataReg) & 0x08)
		Data = 1;
	//end spinlock...
	LEAVE_CRITICAL_SECTION
	return Data;
}